from typing import NamedTuple, Tuple, List

import torch
import torch.nn as nn

from timm.models.vision_transformer import Block
from timm.layers.helpers import to_3tuple


class ViT3dConfig(NamedTuple):
    image_size: Tuple[int, int, int] = (128, 128, 128)
    patch_size: int = 16
    in_channels: int = 1
    embed_dim: int = 768
    depth: int = 12
    num_heads: int = 12
    num_registers: int = 4
    drop_path_rate: float = 0.0


class ViT3dOutput(NamedTuple):
    cls_token: torch.Tensor
    reg_tokens: torch.Tensor
    patch_tokens: torch.Tensor
    intermediate_cls_tokens: List[torch.Tensor]
    intermediate_reg_tokens: List[torch.Tensor]
    intermediate_patch_tokens: List[torch.Tensor]


class PatchEmbed3d(nn.Module):
    def __init__(self, config: ViT3dConfig) -> None:
        super().__init__()

        patch_size = to_3tuple(config.patch_size)

        self.conv = nn.Conv3d(config.in_channels, config.embed_dim, kernel_size=patch_size, stride=patch_size)
        self.ln = nn.LayerNorm(config.embed_dim)
    
    def forward(self, image: torch.Tensor) -> torch.Tensor:
        x = self.conv(image)  # (N, D, H, W, S)
        x = x.flatten(2).movedim(-1, 1)  # (N, H * W * S, D)
        x = self.ln(x)  # (N, H * W * S, D)
        return x


class ViT3d(nn.Module):
    def __init__(self, config: ViT3dConfig) -> None:
        super().__init__()

        self.config = config

        self.image_size = to_3tuple(config.image_size)
        self.patch_size = to_3tuple(config.patch_size)
        self.grid_size = tuple(s // p for s, p in zip(self.image_size, self.patch_size))
        self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        self.num_tokens = 1 + config.num_registers + self.num_patches

        self.patch_embed = PatchEmbed3d(config)
        self.cls_token = nn.Parameter(torch.zeros(1, 1, config.embed_dim))
        self.reg_token = nn.Parameter(torch.zeros(1, config.num_registers, config.embed_dim)) if config.num_registers > 0 else None
        self.pos_embed = nn.Parameter(torch.zeros(1, self.num_tokens, config.embed_dim))
        self.norm_pre = nn.LayerNorm(config.embed_dim)

        drop_path_rates = torch.linspace(0, config.drop_path_rate, config.depth).tolist()
        self.blocks = nn.Sequential(*[
            Block(
                dim=config.embed_dim,
                num_heads=config.num_heads,
                drop_path=drop_path_rates[i],
            )
            for i in range(config.depth)
        ])
        self.norm = nn.LayerNorm(config.embed_dim)

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(image)
        batch_size = x.size(0)

        to_cat = []
        to_cat.append(self.cls_token.expand(batch_size, -1, -1))
        if self.reg_token is not None:
            to_cat.append(self.reg_token.expand(batch_size, -1, -1))
        x = torch.cat(to_cat + [x], dim=1)

        x = x + self.pos_embed

        x = self.norm_pre(x)

        intermediate_cls_tokens, intermediate_reg_tokens, intermediate_patch_tokens = [], [], []
        for block in self.blocks:
            x = block(x)

            cls_token, reg_tokens, patch_tokens = x.split((1, self.config.num_registers, self.num_patches), dim=1)
            patch_tokens = patch_tokens.view(batch_size, *self.grid_size, self.config.embed_dim)

            intermediate_cls_tokens.append(cls_token)
            intermediate_reg_tokens.append(reg_tokens)
            intermediate_patch_tokens.append(patch_tokens)

        x = self.norm(x)

        cls_token, reg_tokens, patch_tokens = x.split((1, self.config.num_registers, self.num_patches), dim=1)
        patch_tokens = patch_tokens.view(batch_size, *self.grid_size, self.config.embed_dim)

        return ViT3dOutput(cls_token, reg_tokens, patch_tokens, intermediate_cls_tokens,
                           intermediate_reg_tokens, intermediate_patch_tokens)



import torch
import torch.nn as nn


class LayerNorm3d(nn.LayerNorm):
    def __init__(self, channels: int, eps: float = 1e-6, affine: bool = True):
        super().__init__(channels, eps=eps, elementwise_affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 3, 4, 1)
        x = super().forward(x)
        x = x.permute(0, 4, 1, 2, 3)
        return x


class GlobalResponseNorm3d(nn.Module):
    def __init__(self, channels: int, eps: float = 1e-6):
        super().__init__()

        self.weight = nn.Parameter(torch.zeros(channels))
        self.bias = nn.Parameter(torch.zeros(channels))
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_g = x.norm(p=2, dim=(2, 3, 4), keepdim=True)
        x_n = x_g / (x_g.mean(dim=1, keepdim=True) + self.eps)
        x = x + torch.addcmul(self.bias.view(-1, 1, 1, 1), self.weight.view(-1, 1, 1, 1), x * x_n)
        return x



from typing import Optional, List, Union, Tuple, Sequence, Any, NamedTuple

import torch
import torch.nn as nn

from timm.layers import DropPath
from timm.layers.helpers import to_3tuple

from medimm.layers.norm import LayerNorm3d, GlobalResponseNorm3d
from medimm.layers.layer_scale import LayerScale3d


class ConvNeXt3dConfig(NamedTuple):
    in_channels: int = 1
    channels: Sequence[int] = (96, 192, 384, 768)
    depths: Sequence[Union[int, Tuple[int, int]]] = (3, 3, 9, 3)
    stem_stride: Union[int, Tuple[int, int, int]] = 4
    stem_kernel_size: Optional[Union[int, Tuple[int, int, int]]] = None
    stem_padding: Union[int, Tuple[int, int, int]] = 0
    drop_path_rate: float = 0.0


class ConvNeXt3dOutput(NamedTuple):
    feature_pyramid: List[torch.Tensor]
    pooled_features: torch.Tensor


class ConvNeXtBlock3d(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            hidden_factor: float = 4.0,
            kernel_size: int = 3,
            dropout_rate: float = 0.0,
            drop_path_rate: float = 0.0,
            grn: bool = False,
            layer_scale: bool = True,
    ) -> None:
        super().__init__()

        hidden_channels = int(in_channels * hidden_factor)
        self.conv_1 = nn.Conv3d(in_channels, in_channels, kernel_size, padding='same', groups=in_channels)
        self.norm = LayerNorm3d(in_channels)
        self.conv_2 = nn.Conv3d(in_channels, hidden_channels, kernel_size=1)
        self.act = nn.GELU()
        self.grn = GlobalResponseNorm3d(hidden_channels) if grn else nn.Identity()
        self.dropout = nn.Dropout(dropout_rate) if dropout_rate > 0 else nn.Identity()
        self.conv_3 = nn.Conv3d(hidden_channels, out_channels, kernel_size=1)
        self.layerscale = LayerScale3d(out_channels, init_values=1e-6) if layer_scale else nn.Identity()
        self.drop_path = DropPath(drop_path_rate)

        if in_channels != out_channels:
            self.shortcut = nn.Conv3d(in_channels, out_channels, kernel_size=1)
        else:
            self.shortcut = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        input_ = x
        x = self.conv_1(x)
        x = self.norm(x)
        x = self.conv_2(x)
        x = self.act(x)
        x = self.grn(x)
        x = self.dropout(x)
        x = self.conv_3(x)
        x = self.layerscale(x)
        x = self.drop_path(x)
        x = x + self.shortcut(input_)
        return x


class ConvNeXtStage3d(nn.Module):
    def __init__(
            self,
            channels: int,
            depth: int,
            drop_path_rates: Optional[Sequence[float]] = None,
            **convnext_block_kwargs: Any
    ) -> None:
        super().__init__()

        if drop_path_rates is None:
            drop_path_rates = [0.0] * depth

        assert len(drop_path_rates) == depth

        self.blocks = nn.Sequential(*[
            ConvNeXtBlock3d(channels, channels, drop_path_rate=drop_path_rates[i], **convnext_block_kwargs)
            for i in range(depth)
        ])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.blocks(x)


class ConvNeXt3d(nn.Module):
    def __init__(self, config: ConvNeXt3dConfig) -> None:
        super().__init__()

        self.config = config

        stem_stride = to_3tuple(config.stem_stride)
        if config.stem_kernel_size is not None:
            stem_kernel_size = to_3tuple(config.stem_kernel_size)
        else:
            stem_kernel_size = stem_stride
        stem_padding = to_3tuple(config.stem_padding)

        drop_path_rates = torch.linspace(0, config.drop_path_rate, sum(config.depths)).split(config.depths)
        drop_path_rates = [dp_rates.tolist() for dp_rates in drop_path_rates]

        self.stem_conv = nn.Conv3d(config.in_channels, config.channels[0], stem_kernel_size, stem_stride, stem_padding)
        self.stem_norm = LayerNorm3d(config.channels[0])
        self.stages = nn.ModuleList([])
        self.norms = nn.ModuleList([])
        self.down_convs = nn.ModuleList([])
        for c_1, c_2, d, dp_rates in zip(config.channels, config.channels[1:], config.depths, drop_path_rates):
            self.stages.append(ConvNeXtStage3d(c_1, d, dp_rates))
            self.norms.append(LayerNorm3d(c_1))
            self.down_convs.append(nn.Conv3d(c_1, c_2, kernel_size=2, stride=2))
        self.stages.append(ConvNeXtStage3d(config.channels[-1], config.depths[-1], drop_path_rates[-1]))
        self.avg_pool = nn.AdaptiveAvgPool3d(output_size=(1, 1, 1))
        self.final_norm = nn.LayerNorm(config.channels[-1], eps=1e-6)

        self.min_input_size = tuple(s * 2 ** len(self.down_convs) for s in stem_stride)

    def forward(self, image: torch.Tensor) -> ConvNeXt3dOutput:
        if any(image.shape[i] < self.min_input_size[i] for i in [-3, -2, -1]):
            raise ValueError(f"Input's spatial size {x.shape[-3:]} is less than {self.min_input_size}.")

        x = self.stem_conv(image)
        x = self.stem_norm(x)

        feature_pyramid = []
        for stage, norm, conv in zip(self.stages, self.norms, self.down_convs):
            x = stage(x)
            feature_pyramid.append(x)
            x = norm(x)
            x = conv(x)

        x = self.stages[-1](x)
        feature_pyramid.append(x)

        pooled_features = self.avg_pool(x).squeeze((2, 3, 4))
        pooled_features = self.final_norm(pooled_features)

        return ConvNeXt3dOutput(feature_pyramid, pooled_features)



from contextlib import contextmanager

import torch.nn as nn


@contextmanager
def eval_mode(module: nn.Module):
    original_mode = module.training

    try:
        module.eval()
        yield module
    finally:
        module.train(original_mode) 



import torch
import torch.nn as nn


class LayerScale3d(nn.Module):
    def __init__(self, channels: int, init_values: float = 1e-5, inplace: bool = False):
        super().__init__()

        self.inplace = inplace
        self.gamma = nn.Parameter(torch.zeros(channels, 1, 1, 1))
        nn.init.constant_(self.gamma, init_values)

    def forward(self, x):
        return x.mul_(self.gamma) if self.inplace else x * self.gamma



from typing import NamedTuple, Optional, List, Union, Tuple, Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F


class UNet3dConfig(NamedTuple):
    in_channels: int = 1
    out_channels: int = 32
    hidden_channels: int = (32, 64, 128, 256, 512, 1024)
    depths: Sequence[int] = (1, 1, 2, 2, 4, 4)


class UNet3dOutput(NamedTuple):
    output: torch.Tensor
    feature_pyramid: List[torch.Tensor]


class UNetBlock3d(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            stride: Union[int, Tuple[int, int, int]] = 1
    ) -> None:
        super().__init__()

        self.conv_1 = nn.Conv3d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1)
        self.norm_1 = nn.InstanceNorm3d(out_channels, affine=True)
        self.act_1 = nn.LeakyReLU(inplace=True)
        self.conv_2 = nn.Conv3d(out_channels, out_channels, kernel_size=3, padding=1)
        self.norm_2 = nn.InstanceNorm3d(out_channels, affine=True)
        self.act_2 = nn.LeakyReLU(inplace=True)

        if in_channels != out_channels:
            self.shortcut = nn.Conv3d(in_channels, out_channels, kernel_size=1, stride=stride)
        else:
            self.shortcut = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        input_ = x
        x = self.conv_1(x)
        x = self.norm_1(x)
        x = self.act_1(x)
        x = self.conv_2(x)
        x = self.norm_2(x)
        x += self.shortcut(input_)
        x = self.act_2(x)
        return x


class UNet3d(nn.Module):
    def __init__(self, config: UNet3dConfig) -> None:
        super().__init__()

        self.encoder_stages = nn.ModuleList([])
        self.decoder_ups = nn.ModuleList([])
        self.decoder_stages = nn.ModuleList([])

        self.encoder_stages.append(
            nn.Sequential(
                UNetBlock3d(config.in_channels + 1, config.hidden_channels[0], stride=1),
                *[UNetBlock3d(config.hidden_channels[0], config.hidden_channels[0]) for _ in range(config.depths[0] - 1)]
            )
        )
        for i in range(len(config.hidden_channels) - 1):
            self.encoder_stages.append(
                nn.Sequential(
                    UNetBlock3d(config.hidden_channels[i], config.hidden_channels[i + 1], stride=2),
                    *[UNetBlock3d(config.hidden_channels[i + 1], config.hidden_channels[i + 1]) for _ in range(config.depths[i + 1] - 1)]
                )
            )
            self.decoder_ups.append(
                nn.Sequential(
                    nn.Upsample(scale_factor=2, mode='nearest'),
                    nn.Conv3d(config.hidden_channels[i + 1], config.hidden_channels[i], kernel_size=1),
                )
            )
            self.decoder_stages.append(
                nn.Sequential(
                    UNetBlock3d(config.hidden_channels[i] * 2, config.hidden_channels[i]),
                    *[UNetBlock3d(config.hidden_channels[i], config.hidden_channels[i]) for _ in range(config.depths[i] - 1)]
                )
            )
        self.final_conv = nn.Conv3d(config.hidden_channels[0], config.out_channels, kernel_size=1)

    def forward(self, image: torch.Tensor, mask: Optional[torch.Tensor] = None) -> UNet3dOutput:
        if any(image.shape[i] < 2 ** (len(self.encoder_stages) - 1) for i in [-3, -2, -1]):
            raise ValueError(f"Input's spatial size {x.shape[-3:]} is less than {self.max_stride}.")

        if mask is None:
            n, _, h, w, d = image.shape
            mask = torch.ones((n, h, w, d), dtype=image.dtype, device=image.device)
        elif mask.dtype != image.dtype:
            raise TypeError("``mask`` must have the same dtype as input image ``x``")
        mask = mask.unsqueeze(1)
        x = torch.cat([image * mask, mask], dim=1)

        # encoder
        feature_pyramid = []
        for stage in self.encoder_stages:
            x = stage(x)
            feature_pyramid.append(x)

        # decoder
        for i in reversed(range(len(self.decoder_stages))):
            x = self.decoder_ups[i](x)
            y = feature_pyramid[i]
            x = crop_and_pad_to(x, y)
            x = torch.cat([x, y], dim=1)
            x = self.decoder_stages[i](x)
            feature_pyramid[i] = x

        # final conv
        output = self.final_conv(x)

        return UNet3dOutput(output, feature_pyramid)


def crop_and_pad_to(x: torch.Tensor, other: torch.Tensor, pad_mode: str = 'replicate') -> torch.Tensor:
    assert x.ndim == other.ndim == 5

    if x.shape == other.shape:
        return x

    # crop
    x = x[(..., *map(slice, other.shape[-3:]))]

    # pad
    pad = []
    for dim in [-1, -2, -3]:
        pad += [0, max(other.shape[dim] - x.shape[dim], 0)]
    x = F.pad(x, pad, mode=pad_mode)

    return x


# class UNet3dWithConvNextBackbone:
#     """After UNet3d encoder, enrich feature pyramid with ConvNext features.
#     """
#     pass


# class UNet3dWithViTBackbone:
#     """After UNet3d encoder, enrich feature pyramid with ViT features.
#     """
#     pass
